#
#    CassaFS
#    Copyright (C) 2012  Paul Dwerryhouse <paul@dwerryhouse.com.au>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pycassa.pool import ConnectionPool
from pycassa.columnfamily import ColumnFamily
from pycassa.cassandra.ttypes import NotFoundException
from pycassa.types import *
from time import time
from collections import OrderedDict
import uuid, sys
import os.path

try:
    import _find_fuse_parts
except ImportError:
    pass
import fuse
from fuse import Fuse

class NameTooLongException(Exception):
    pass

KEYSPACE="CassaFS"
SERVER="127.0.0.1:9160"

POOL = ConnectionPool(KEYSPACE, server_list=['127.0.0.1:9160'])

DIRECTORY = ColumnFamily(POOL, 'Directory')
INODE = ColumnFamily(POOL, 'Inode')
BLOCK = ColumnFamily(POOL, 'Block')

INODE.column_validators['nlink'] = IntegerType()
INODE.column_validators['mode'] = IntegerType()
INODE.column_validators['rdev'] = IntegerType()
INODE.column_validators['uid'] = IntegerType()
INODE.column_validators['gid'] = IntegerType()
INODE.column_validators['size'] = IntegerType()
INODE.column_validators['blksize'] = IntegerType()
INODE.column_validators['blocks'] = IntegerType()
INODE.column_validators['mtime'] = IntegerType()
INODE.column_validators['ctime'] = IntegerType()
INODE.column_validators['atime'] = IntegerType()

INODE.key_validation_class = LexicalUUIDType()
BLOCK.key_validation_class = LexicalUUIDType()
DIRECTORY.key_validation_class = LexicalUUIDType()

DIRECTORY.default_validation_class = LexicalUUIDType()

BLOCK.column_name_class = IntegerType()

rootkey = uuid.UUID('{00000000-0000-0000-0000-000000000000}')

if not hasattr(fuse, '__version__'):
    raise RuntimeError, \
        "your fuse-py doesn't know of fuse.__version__, probably it's too old."

fuse.fuse_python_api = (0, 2)

def get_key(path):
    key = None
    for p in path.split('/'):
        if p == '':
            key = rootkey
        else:
            try:
                key = DIRECTORY.get(key, columns=[p])[p]
            except KeyError:
                raise NotFoundException
    return key

def get_inode(key):
    try:
        inode = INODE.get(key)
    except NotFoundException:
        return None

    return inode

def set_inode(key, inode):
    INODE.insert(key, inode)

def remove_inode(key):
    INODE.remove(key)

def get_blocks(key, start, finish):
    return BLOCK.get(key, column_start=start, column_finish=finish)

def remove_block(key):
    BLOCK.remove(key)

def remove_blocks(key, start, end):
    BLOCK.remove(key, range(start,end))

def set_block(key, blkid, newblock):
    BLOCK.insert(key, { blkid : newblock })

def insert_dir(key, direntry):
    DIRECTORY.insert(key, direntry)

def get_all_dirs(key):
    return DIRECTORY.get(key,column_count=sys.maxint)

def get_dir(key, filename):
    return DIRECTORY.get(key, columns=[filename])[filename]

def remove_dir(key, filename):
    return DIRECTORY.remove(key, { filename })

def remove_direntry(key):
    return DIRECTORY.remove(key)

def utimens(path, atime, mtime):
    key = get_key(path)
    inode = {}
    inode['atime'] = atime.tv_sec
    inode['mtime'] = mtime.tv_sec
    inode = INODE.insert(key,inode)

def create(dirname,basename,inode):
    key = uuid.uuid4()
    parentkey = get_key(dirname)
    INODE.insert(key,inode)
    DIRECTORY.insert(parentkey, {basename : key})
    return 0
