#!/usr/bin/env python
#
#    CassaFS
#    Copyright (C) 2012  Paul Dwerryhouse <paul@dwerryhouse.com.au>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pycassa.cassandra.ttypes import NotFoundException

from time import time
from collections import OrderedDict
import grp, pwd
import uuid, sys

import os, stat, errno, os.path
try:
    import _find_fuse_parts
except ImportError:
    pass
import fuse
from fuse import Fuse

import cassafs.cass as cass

class NameTooLongException(Exception):
    pass

BLKSIZE = 8192

if not hasattr(fuse, '__version__'):
    raise RuntimeError, \
        "your fuse-py doesn't know of fuse.__version__, probably it's too old."

fuse.fuse_python_api = (0, 2)


def replace(s,buf,p):
    ns = ""
    n = len(s)
    bn = len(buf)
    if n <= p:
        ns = s + ('\0' * (p-n)) + buf
    else:
        ns = s[0:p] + buf + s[p+bn:]
    return ns

def writeblock(key, buf, blkid, pos):
    try:
        block = cass.get_blocks(key, blkid, blkid)
    except NotFoundException:
        block = OrderedDict([(blkid, b'')])
    
    newblock = replace(block[blkid],buf,pos)
    cass.set_block(key, blkid, newblock)

    return len(buf)

def check_filename_len(path):
    for p in path.split('/'):
        if len(p) > 255:
            raise NameTooLongException

def user_in_group(uid,gid):
    
    try:
        if pwd.getpwuid(uid).pw_gid == gid:
            return True
    except KeyError:
        pass

    try:
        for user in grp.getgrgid(gid).gr_mem:
            if uid == pwd.getpwnam(user).pw_uid:
                return True
    except KeyError:
        pass

    return False
    
def get_path_components(path):
    paths = [ path ]
    while path != '/':
        path = os.path.dirname(path)
        paths.insert(0,path)
    try:
        end = paths.pop()
    except IndexError:
        end = ''

    try:
        notend = paths.pop()
    except IndexError:
        notend = ''
    
    return (end, notend, paths)
    
class CassaFSStat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0

class CassaFS(Fuse):

    def getattr(self, path):
        st = CassaFSStat()
        try:
            key = cass.get_key(path)
            inode = cass.get_inode(key)
            if inode.get('mode') != None:
                st.st_mode = inode.get('mode')
            if inode.get('nlink') != None:
                st.st_nlink = inode.get('nlink')
            if inode.get('rdev') != None:
                st.st_rdev = inode.get('rdev')
            if inode.get('uid') != None:
                st.st_uid = inode.get('uid')
            if inode.get('gid') != None:
                st.st_gid = inode.get('gid')
            if inode.get('atime') != None:
                st.st_atime = inode.get('atime')
            if inode.get('mtime') != None:
                st.st_mtime = inode.get('mtime')
            if inode.get('ctime') != None:
                st.st_ctime = inode.get('ctime')
            if inode.get('blocks') != None:
                st.st_blocks = inode.get('blocks')
            if inode.get('size') != None:
                st.st_size = inode.get('size')
        except NotFoundException:
            return -errno.ENOENT

        return st

    def check_access(self, path, mode):

        try:
            key = cass.get_key(path)
        except NotFoundException:
            return -errno.EEXIST

        if mode == os.F_OK:
            return 0

        inode = cass.get_inode(key)
        fmode = inode['mode']
        context = self.GetContext()
        uid = context['uid']
        gid = context['gid']


        if (inode['uid'] == uid):
            if (mode & os.X_OK) and (fmode & stat.S_IXUSR):
                return 0

            if (mode & os.W_OK) and (fmode & stat.S_IWUSR):
                return 0

            if (mode & os.R_OK) and (fmode & stat.S_IRUSR):
                return 0

        if (inode['gid'] == gid):
            if (mode & os.X_OK) and (fmode & stat.S_IXGRP):
                return 0

            if (mode & os.W_OK) and (fmode & stat.S_IWGRP):
                return 0

            if (mode & os.R_OK) and (fmode & stat.S_IRGRP):
                return 0

        if (mode & os.X_OK) and (fmode & stat.S_IXOTH):
            return 0

        if (mode & os.W_OK) and (fmode & stat.S_IWOTH):
            return 0

        if (mode & os.R_OK) and (fmode & stat.S_IROTH):
            return 0
        
        return -errno.EACCES

    
    def access(self, path, mode):
        (first, second, paths) = get_path_components(path)

        # Check that we can access leading paths
        for p in paths:
            err = self.check_access(p, os.X_OK)
            if err < 0:
                return err

        if second != '':
            err = self.check_access(second, os.X_OK)
            if err < 0:
                return err

        err = self.check_access(first, mode)
        if err < 0:
            return err

        return 0


    def readdir(self, path, offset):
        # XXX currently ignoring the offset
        key = cass.get_key(path)
        dirs = cass.get_all_dirs(key)
        for r in dirs:
            yield fuse.Direntry(r)

    def chown(self, path, uid, gid):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)

            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err

            if second != '':
                err = self.check_access(second, os.X_OK)
                if err < 0:
                    return err

        key = cass.get_key(path)
        inode = cass.get_inode(key)

        if gid == -1:
            if context['uid'] != 0:
                return -errno.EPERM

            inode = {
                'uid': uid,
            }

        elif uid == -1:
            if context['uid'] != 0:
                if context['uid'] != inode['uid']:
                    return -errno.EPERM

                if context['gid'] != gid and not user_in_group(context['uid'],gid):
                    return -errno.EPERM

            inode = {
                'gid': gid,
            }

        else:
            if context['uid'] != inode['uid'] and context['uid'] != 0:
                return -errno.EPERM

            if context['uid'] != 0 and context['uid'] != uid:
                return -errno.EPERM
            
            inode = {
                'uid': uid,
                'gid': gid,
            }

        now = int(time())
        inode['ctime'] = now
        cass.set_inode(key,inode)
        

    def chmod(self, path, mode):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK)
                if err < 0:
                    return err

        key = cass.get_key(path)
        inode = cass.get_inode(key)

        if context['uid'] != inode['uid'] and context['uid'] != 0:
            return -errno.EPERM
    
        oldmode = inode.get('mode')
        mode = (oldmode & 0770000) | ( mode & 07777)
        if context['uid'] != 0 and context['gid'] != inode.get('gid'):
            mode = mode & 0775777
        now = int(time())
        inode = {
            'mode': mode,
            'ctime': now
        }
        cass.set_inode(key,inode)
        
    def mknod(self, path, mode, dev):
        try:
            check_filename_len(path)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        key = uuid.uuid4()
        now = int(time())

        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = cass.get_key(dirname)

        inode = {
            'uid': 0,
            'gid': 0,
            'rdev': dev,
            'mode': mode,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'nlink': 1,
            'atime': now,
            'mtime': now,
            'ctime': now,
        }
        cass.set_inode(key,inode)
        cass.insert_dir(parentkey, {basename : key})

    def rmdir(self, path):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = cass.get_key(path)
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = cass.get_key(dirname)
        dirs = cass.get_all_dirs(key)
        if len(dirs) > 2:
            return -errno.ENOTEMPTY
        
        cass.remove_dir(parentkey, basename)
        cass.remove_direntry(key)
        cass.remove_inode(key)
        now = int(time())
        inode = { 'mtime': now,
                  'ctime': now }
        cass.set_inode(parentkey,inode)
    
        return 0

    def mkdir(self, path, mode):
        try:
            check_filename_len(path)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = uuid.uuid4()
        now = int(time())
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = cass.get_key(dirname)
        newdir = { basename: key }

        thisdir = { '.': key }
        parentdir = { '..': parentkey }

        inode = {
            'mode': mode | stat.S_IFDIR,
            'uid': context['uid'],
            'gid': context['gid'],
            'rdev': 0,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'nlink': 2,
            'atime': now,
            'mtime': now,
            'ctime': now,
        }
        cass.set_inode(key,inode)
        cass.insert_dir(key,thisdir)
        cass.insert_dir(key,parentdir)
        cass.insert_dir(parentkey, newdir)
        
        return 0

    def rename(self, path1, path2):
        try:
            check_filename_len(path2)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        key1 = cass.get_key(path1)
        inode1 = cass.get_inode(key1)

        file1 = os.path.basename(path1)
        dir1 = os.path.dirname(path1)
        err = self.access(dir1, os.W_OK)
        if err < 0:
            return err

        file2 = os.path.basename(path2)
        dir2 = os.path.dirname(path2)
        err = self.access(dir2, os.W_OK)
        if err < 0:
            return err

        dirkey1 = cass.get_key(dir1)
        direntry = cass.get_dir(dirkey1, file1)
        cass.remove_dir(dirkey1, file1)
        dirkey2 = cass.get_key(dir2)
        newdirentry = { file2: key1 }
        cass.insert_dir(dirkey2, newdirentry)
        
    def unlink(self, path):
        (first, second, paths) = get_path_components(path)
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = cass.get_key(path)
        inode = cass.get_inode(key)
            
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = cass.get_key(dirname)
        cass.remove_dir(parentkey, basename)
        if inode['nlink'] <= 1:
            cass.remove_inode(key)
            cass.remove_block(key)
        else:
            inode['nlink']-=1
            cass.set_inode(key,inode)
        
    def readlink(self, path):
        key = cass.get_key(path)
        inode = cass.get_inode(key)
        return inode['symlink']

    def symlink(self, targetpath, linkpath):
        try:
            check_filename_len(linkpath)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(linkpath)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = uuid.uuid4()
        now = int(time())
        basename = os.path.basename(linkpath)
        dirname = os.path.dirname(linkpath)
        parentkey = cass.get_key(dirname)
        inode = {
            'mode': 777 | stat.S_IFLNK,
            'uid': context['uid'],
            'gid': context['gid'],
            'nlink': 1,
            'rdev': 0,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'atime': now,
            'mtime': now,
            'ctime': now,
            'symlink': targetpath
        }
        cass.set_inode(key,inode)
        cass.insert_dir(parentkey, {basename : key})

    def link(self, targetpath, linkpath):
        try:
            check_filename_len(linkpath)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        context = self.GetContext()
        key = cass.get_key(targetpath)
        inode = cass.get_inode(key)

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(targetpath)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

            if inode['uid'] != context['uid']:
                return -errno.EACCESS

        basename = os.path.basename(linkpath)
        dirname = os.path.dirname(linkpath)
        parentkey = cass.get_key(dirname)
        inode['nlink'] += 1
        cass.set_inode(key,inode)
        cass.insert_dir(parentkey, {basename : key})

    def truncate(self, path, size):
        try:
            check_filename_len(path)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        key = cass.get_key(path)
        inode = cass.get_inode(key)
        total = inode['size']
        blkid = int(size / BLKSIZE)
        endblkid = int(total / BLKSIZE)
        cass.remove_blocks(key, blkid+1, endblkid+1)
        offset = size % BLKSIZE

        try:
            block = cass.get_blocks(key,blkid,blkid)
        except NotFoundException:
            block = OrderedDict([(blkid, b'')])

        newblock = block[blkid][0:offset]
        cass.set_block(key, blkid, newblock)
        cass.set_inode(key, { 'size': size })

    def utimens(self, path, atime, mtime):
        return cass.utimens(path,atime,mtime)

    def create(self, path, flags, mode):
        try:
            check_filename_len(path)
        except NameTooLongException:
            return -errno.ENAMETOOLONG

        (first, second, paths) = get_path_components(path)

        # Check that we can access leading paths
        for p in paths:
            err = self.check_access(p, os.X_OK)
            if err < 0:
                return err

        if second != '':
            err = self.check_access(second, os.X_OK | os.W_OK)
            if err < 0:
                return err

        err = self.check_access(first, mode)
        if err < 0 and err != -errno.EEXIST:
            return err

        context = self.GetContext()
        key = uuid.uuid4()
        now = int(time())
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = cass.get_key(dirname)
        inode = {
            'mode': mode | stat.S_IFREG,
            'uid': context['uid'],
            'gid': context['gid'],
            'nlink': 1,
            'rdev': 0,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'atime': now,
            'mtime': now,
            'ctime': now,
        }
        return cass.create(dirname,basename,inode)
        

    def open(self, path, flags):

        # XXX maybe should look to see if O_TRUNC/atomic_o_trunc causes
        # any concern here

        if flags & 0x03 == os.O_RDONLY:
            return self.access(path, os.R_OK)

        elif flags & 0x03 == os.O_WRONLY:
            return self.access(path, os.W_OK)

        elif flags & 0x03 == os.O_RDWR:
            return self.access(path, os.R_OK | os.W_OK)


    def write(self, path, buf, offset):

        key = cass.get_key(path)
        buflen = len(buf)
        n = buflen
        start = int(offset / BLKSIZE)
        finish = int((offset + n - 1) / BLKSIZE)
        inode = cass.get_inode(key)
        size = inode['size']

        o = offset % BLKSIZE
        p = 0

        print "WRITE {0} bytes at offset {1}".format(buflen,offset)
        for i in range(start,finish+1):
            if (n <= BLKSIZE - o):
                writeblock(key, buf[p:], start, o)
                break
            l = BLKSIZE - o
            writeblock(key, buf[p:p+l], i, o)
            p += l
            n -= l
            o = 0
            
        if (offset + buflen) > size:
            size = offset + buflen
            inode = { 'size': offset + buflen }

        inode['blocks'] = (offset % BLKSIZE) + 1
        cass.set_inode(key,inode)
        
        return n

    def read(self, path, size, offset):

        key = cass.get_key(path)
        blkid = 0
        start = int(offset / BLKSIZE)
        finish = int((offset + size - 1) / BLKSIZE)
        try:
            block = cass.get_blocks(key, start, finish)
        except NotFoundException:
            return ''
        
        buf = ''
        o = offset % BLKSIZE
        for i in range(start,finish+1):
            if (size <= BLKSIZE - o):
                buf += block[i][o:o+size]
                break
            buf += block[i][o:]
            size -= len(block[i][o:])
            o = 0
            
        return buf

    # XXX this is all clearly bogus
    def statfs(self):

        fs = fuse.StatVfs()
        fs.f_namemax = 255
        fs.f_bsize = BLKSIZE
        fs.f_frsize = BLKSIZE
        fs.f_blocks = 0
        fs.f_bfree = 0
        fs.f_bavail = 0

        return fs

def main():
    usage="""
Userspace CassaFS example

""" + Fuse.fusage
    server = CassaFS(version="%prog " + fuse.__version__,
                     usage=usage,
                     dash_s_do='setsingle')

    server.parse(errex=1)
    server.main()

if __name__ == '__main__':
    main()
